/*
 * Creator: Naman Dixit
 * Notice: © Copyright 2020 Naman Dixit
 */

#include "nlib/nlib.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/socket.h>
#include <sys/epoll.h>
#include <netdb.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/time.h>
#include <arpa/inet.h>

#define MAX_SOCKET_CONNECTIONS_REQUEST 64
#define MAX_EPOLL_EVENTS 1024

typedef struct Input_Resume {
    Char *buffer;
    Size buffer_len;
    Size buffer_cap;
    U32 buffer_expected_len;

    Byte size_bytes[4];
    Size size_bytes_count;

    int fd;
    B32 initialized;
} Input_Resume;

typedef struct Output_Resume {
    Char *buffer;
    Size buffer_pos;
    Size buffer_len;

    U64 msg_id_hash;

    Sint fd;

    Byte _pad[4];
} Output_Resume;

typedef struct Input_Output {
    Input_Resume *ir;
    Output_Resume *ors;
} Input_Output;

typedef struct Grunt {
    Char *ip;
    Sint memory;
} Grunt;

typedef struct Grunt_Survey {
    Char **ips;
    U64 milli_passed;
    U64 milli_last;
    Sint id;
    Sint dispatcher_socket;
} Grunt_Survey;

typedef struct Command {
    enum Command_Kind {
        Command_NONE,
        Command_REQUEST_EXTERNAL,
        Command_REQUEST_INTERNAL,
        Command_RESPONSE_INTERNAL,
        Command_RESPONSE_EXTERNAL,
        Command_HEARTBEAT,
    } kind;

    Sint id;

    union {
        struct {
            Sint memory;
        } reqex;

        struct {
            Sint memory;
            Sint dispatcher_socket;
        } reqin;

        struct {
            Char **ips;
            Sint dispatcher_socket;
        } resin;

        struct {
            Char **ips;
            Sint dispatcher_socket;
        } resex;

        struct {
            Sint memory;
        } heartbeat;
    };
} Command;

# if defined(COMPILER_CLANG)
#  pragma clang diagnostic push
#   pragma clang diagnostic ignored "-Wpadded"
#   pragma clang diagnostic ignored "-Wfloat-equal"
# endif
#include "cJSON/cJSON.h"
#include "cJSON/cJSON.c"
# if defined(COMPILER_CLANG)
#  pragma clang diagnostic pop
# endif

#include "socket.c"
#include "time.c"

Sint main (Sint argc, Char *argv[])
{
    unused_variable(argc);
    unused_variable(argv);

    Sint sock_fd_external = socketCreateListener("9526");
    Sint sock_fd_internal = socketCreateListener("9527");

    // NOTE(naman): Create an epoll instance, with no flags set.
    int epoll_fd = epoll_create1(0);
    if (epoll_fd < 0) {
        perror("epoll_create1");
        exit(-1);
    }

    // NOTE(naman): Add the socket's file descriptor to the epoll set
    struct epoll_event accept_event_internal = {.data.fd = sock_fd_internal,
                                                .events = EPOLLIN};
    if (epoll_ctl(epoll_fd, EPOLL_CTL_ADD, sock_fd_internal, &accept_event_internal) < 0) {
        perror("epoll_ctl EPOLL_CTL_ADD");
        exit(-1);
    }

    struct epoll_event accept_event_external = {.data.fd = sock_fd_external,
                                                .events = EPOLLIN};
    if (epoll_ctl(epoll_fd, EPOLL_CTL_ADD, sock_fd_external, &accept_event_external) < 0) {
        perror("epoll_ctl EPOLL_CTL_ADD");
        exit(-1);
    }

    // NOTE(naman): Allocate memory for epoll array
    struct epoll_event* events = calloc(MAX_EPOLL_EVENTS, sizeof(struct epoll_event));
    if (events == NULL) {
        fprintf(stderr, "Unable to allocate memory for epoll_events");
        exit(-1);
    }

    Command  *commands  = NULL;
    Hash_Table commands_pending = htCreate(0);

    Hash_Table socket_map = htCreate(0);
    Hash_Table io_map = htCreate(0);
    Hash_Table grunt_map = htCreate(0);
    Hash_Table grunt_survey_map = htCreate(0);

    while (true) {
        // NOTE(naman): Get the fd's that are ready
        int nready = epoll_wait(epoll_fd, events, MAX_EPOLL_EVENTS, 100);

        for (int i = 0; i < nready; i++) {
            if (events[i].events & EPOLLERR) {
                perror("epoll_wait returned EPOLLERR");
                exit(-1);
            }

            if ((events[i].data.fd == sock_fd_internal) ||
                (events[i].data.fd == sock_fd_external)) {
                Sint sock_fd = events[i].data.fd;

                // NOTE(naman): A new grunt is connecting.
                struct sockaddr_in peer_addr = {0};
                socklen_t peer_addr_len = sizeof(peer_addr);
                int accept_fd = accept(sock_fd, (struct sockaddr*)&peer_addr,
                                       &peer_addr_len);

                if (accept_fd < 0) {
                    if (errno == EAGAIN || errno == EWOULDBLOCK) {
                        // This can happen due to the nonblocking socket mode; in this
                        // case don't do anything, but print a notice (since these events
                        // are extremely rare and interesting to observe...)
                        fprintf(stderr, "accept() returned %s\n",
                                errno == EAGAIN ? "EAGAIN" : "EWOULDBLOCK");
                    } else {
                        perror("accept");
                        exit(-1);
                    }
                } else {
                    printf("Log: Connection made: client_fd=%d\n", accept_fd);

                    // NOTE(naman): Set the socket as non-blocking
                    int sock_flags = fcntl(sock_fd, F_GETFL, 0);
                    if (sock_flags == -1) {
                        perror("fcntl F_GETFL");
                        exit(-1);
                    }

                    if (fcntl(sock_fd, F_SETFL, sock_flags | O_NONBLOCK) == -1) {
                        perror("fcntl F_SETFL O_NONBLOCK");
                        exit(-1);
                    }

                    // NOTE(naman): Add the new file descriptor to the epoll set
                    struct epoll_event event = {.data.fd = accept_fd,
                                                .events = EPOLLIN};
                    if (epoll_ctl(epoll_fd, EPOLL_CTL_ADD, accept_fd, &event) < 0) {
                        perror("epoll_ctl EPOLL_CTL_ADD");
                        exit(-1);
                    }

                    if (events[i].data.fd == sock_fd_internal) {
                        htInsert(&socket_map, (U64)accept_fd, Socket_Kind_INTERNAL);

                        struct in_addr ip_addr = peer_addr.sin_addr;
                        char str[INET_ADDRSTRLEN];
                        inet_ntop(AF_INET, &ip_addr, str, INET_ADDRSTRLEN);

                        Grunt *grunt = calloc(1, sizeof(*grunt));
                        grunt->ip = strdup(str);
                        htInsert(&grunt_map, (U64)accept_fd, (U64)grunt);
                    } else if (events[i].data.fd == sock_fd_external) {
                        htInsert(&socket_map, (U64)accept_fd, Socket_Kind_EXTERNAL);
                    }

                    Input_Output *io = calloc(1, sizeof(*io));
                    htInsert(&io_map, (U64)accept_fd, (Uptr)io);
                }
            } else {
                // A peer socket is ready.
                if (events[i].events & EPOLLIN) {
                    // Ready for reading.
                    int fd = events[i].data.fd;

                    Socket_Kind socket_kind = (Socket_Kind)htLookup(&socket_map, (U64)fd);

                    Input_Output *io = (Input_Output *)htLookup(&io_map, (U64)fd);


                    if (io->ir == NULL) {
                        io->ir = calloc(1, sizeof(*io->ir));
                        io->ir->buffer_cap = MiB(1);
                        io->ir->buffer = calloc(io->ir->buffer_cap,
                                                    sizeof(*(io->ir->buffer)));
                        io->ir->fd = fd;
                    }

                    Input_Resume *ir = io->ir;

                    if (ir->initialized == false) {
                        long len = read(fd,
                                        (Char *)ir->size_bytes + ir->size_bytes_count,
                                        4 - ir->size_bytes_count);

                        if (len == 0) {
                            close(fd);
                            sbufDelete(io->ors);
                            free(io->ir);
                            free(io);

                            htRemove(&io_map, (U64)fd);
                            htRemove(&socket_map, (U64)fd);

                            if (socket_kind == Socket_Kind_INTERNAL) {
                                htRemove(&grunt_map, (U64)fd);
                            }

                            continue;
                        }

                        ir->size_bytes_count += (Size)len;

                        if (ir->size_bytes_count == 4) {
                            ir->initialized = true;

                            ir->buffer_expected_len = (U32)((ir->size_bytes[3] <<  0U) |
                                                            (ir->size_bytes[2] <<  8U) |
                                                            (ir->size_bytes[1] << 16U) |
                                                            (ir->size_bytes[0] << 24U));
                        }

                        continue;
                    } else {
                        long len = read(fd,
                                        ir->buffer + ir->buffer_len,
                                        ir->buffer_expected_len - ir->buffer_len);

                        ir->buffer_len += (Size)len;

                        if (ir->buffer_expected_len == ir->buffer_len) {
//                            char *json_printed = cJSON_Print(cJSON_Parse(ir->buffer));

                            const Char *json_error = NULL;
                            cJSON *root = cJSON_ParseWithOpts(ir->buffer, &json_error, true);

                            B32 prepare_for_output = true;

                            if (root == NULL) {
                                // TODO(naman): Error
                            } else {
                                if (socket_kind == Socket_Kind_EXTERNAL) {
                                    printf("Recieved: REQUEST EXTERNAL:\n%s\n",
                                           cJSON_Print(cJSON_Parse(ir->buffer)));

                                    Command c = {.kind = Command_REQUEST_INTERNAL};
                                    c.id = cJSON_GetObjectItem(root, "id")->valueint;

                                    Sint memory = cJSON_GetObjectItem(root, "memory")->valueint;

                                    Char **ips = NULL;

                                    for (Size j = 0; j < grunt_map.slot_count; j++) {
                                        if (grunt_map.keys[j] != 0) {
                                            Grunt *g = (Grunt *)grunt_map.values;
                                            if (g->memory >= memory) {
                                                c.kind = Command_RESPONSE_EXTERNAL;
                                                sbufAdd(ips, g->ip);
                                            }
                                        }
                                    }

                                    if (c.kind == Command_REQUEST_INTERNAL) {
                                        c.reqin.dispatcher_socket = fd;
                                        c.reqin.memory = memory;
                                    } else if (c.kind == Command_RESPONSE_EXTERNAL) {
                                        c.resex.ips = ips;
                                        c.reqin.dispatcher_socket = fd;
                                    }

                                    sbufAdd(commands, c);
                                    htInsert(&commands_pending, hashInteger((U64)c.id), true);
                                } else if (socket_kind == Socket_Kind_INTERNAL) {
                                    Char *type = cJSON_GetObjectItem(root, "type")->valuestring;

                                    if (strcmp(type, "response") == 0) {
                                        printf("Recieved: RESPONSE INTERNAL:\n%s\n",
                                               cJSON_Print(cJSON_Parse(ir->buffer)));

                                        B32 success = (B32)(cJSON_GetObjectItem(root,
                                                                                "success")->valueint);

                                        if (success) {
                                            Command c = {.kind = Command_RESPONSE_EXTERNAL};
                                            c.id = cJSON_GetObjectItem(root, "id")->valueint;

                                            c.resex.dispatcher_socket = cJSON_GetObjectItem(root,
                                                                                      "dispatcher_socket")->valueint;

                                            Grunt_Survey *gs =
                                                (Grunt_Survey *)htLookup(&grunt_survey_map,
                                                                         hashInteger((U64)c.id));

                                            if (gs != NULL) {
                                                Grunt *g = (Grunt *)htLookup(&grunt_map, (U64)fd);

                                                sbufAdd(gs->ips, g->ip);

                                                U64 milli_new = timeMilli();
                                                gs->milli_passed += milli_new - gs->milli_last;
                                                gs->milli_last = milli_new;

                                                if (gs->milli_passed >= 1000) {
                                                    htRemove(&grunt_survey_map,
                                                             hashInteger((U64)c.id));

                                                    c.resex.ips = gs->ips;
                                                    sbufAdd(commands, c);
                                                }
                                            }
                                        }
                                    } else if (strcmp(type, "heartbeat") == 0) {
                                        printf("Recieved: HEARTBEAT:\n%s\n",
                                               cJSON_Print(cJSON_Parse(ir->buffer)));

                                        Grunt *grunt = (Grunt *)htLookup(&grunt_map, (U64)fd);
                                        grunt->memory = cJSON_GetObjectItem(root, "memory")->valueint;

                                        prepare_for_output = false;
                                    }
                                }
                            }

                            free(ir->buffer);
                            free(ir);
                            io->ir = NULL;

                            if (prepare_for_output) {
                                struct epoll_event event = {.data.fd = fd,
                                                            .events = EPOLLIN | EPOLLOUT};
                                if (epoll_ctl(epoll_fd, EPOLL_CTL_MOD, fd, &event) < 0) {
                                    perror("epoll_ctl EPOLL_CTL_MOD");
                                    exit(-1);
                                }
                            }
                        }
                    }
                } else if (events[i].events & EPOLLOUT) {
                    // Writing into fd in which we previously were not able to finish writing to
                    int fd = events[i].data.fd;

                    Input_Output *io = (Input_Output *)htLookup(&io_map, (U64)fd);

                    if (sbufElemin(io->ors) == 0) {
                        // fprintf(stderr, "hmLookup returned NULL\n");
                        // NOTE(naman): We haven't popped the Result yet, first go do that.
                        continue;
                    } else {
                        Output_Resume *or = &io->ors[0];

                        Char *output = or->buffer;
                        Size output_len = or->buffer_len;
                        Size output_pos = or->buffer_pos;

                        ssize_t nsent = write(or->fd, output + output_pos, output_len - output_pos);

                        if (nsent == -1) {
                            if (errno == EAGAIN || errno == EWOULDBLOCK) {
                                // Try next time
                            } else {
                                perror("write() failed\n");
                                exit(-1);
                            }
                        } else if ((Size)nsent < (output_len - output_pos)) {
                            or->buffer_pos += (Size)nsent;
                        } else {
                            sbufUnsortedRemove(io->ors, 0);

                            struct epoll_event event = {.data.fd = fd,
                                                        .events = EPOLLIN};
                            if (epoll_ctl(epoll_fd, EPOLL_CTL_MOD,
                                          or->fd, &event) < 0) {
                                perror("epoll_ctl EPOLL_CTL_MOD");
                                exit(-1);
                            }

                            sbufDelete(output);
                        }
                    }
                }
            }
        }

        for (Size i = 0; i < grunt_survey_map.slot_count; i++) {
            if (grunt_survey_map.keys[i] != 0) {
                Grunt_Survey *gs = (Grunt_Survey *)grunt_survey_map.values[i];

                U64 milli_new = timeMilli();
                gs->milli_passed += milli_new - gs->milli_last;
                gs->milli_last = milli_new;

                if (gs->milli_passed >= 1000) {
                    htRemove(&grunt_survey_map,
                             hashInteger((U64)gs->id));

                    Command c = {.kind = Command_RESPONSE_EXTERNAL};
                    c.id = gs->id;
                    c.resex.dispatcher_socket = gs->dispatcher_socket;
                    c.resex.ips = gs->ips;

                    sbufAdd(commands, c);
                }
            }
        }

        for (Size j = 0; j < sbufElemin(commands); j++) {
            Command c = commands[j];

            if (c.kind == Command_REQUEST_INTERNAL) {
                for (Size i = 0; i < socket_map.slot_count; i++) {
                    if ((socket_map.values[i] == Socket_Kind_INTERNAL) &&
                        (socket_map.keys[i] != 0)) {
                        Char *output = NULL;

                        sbufPrint(output, "    ");
                        sbufPrint(output, "{\n\"id\": %d", c.id);
                        sbufPrint(output, ",\n\"dispatcher_socket\": %d", c.reqin.dispatcher_socket);

                        sbufPrint(output, ",\n\"memory\": %d\n", c.reqin.memory);

                        sbufPrint(output, "\n}\n");

                        printf("Sending: REQUEST INTERNAL:\n%s\n",
                               cJSON_Print(cJSON_Parse(output + 4)));

                        Size output_len = strlen(output);

#if defined(ENDIAN_LITTLE)
                        U32 json_len = (U32)output_len - 4;
                        U32 json_len_be = swap_endian(json_len);
                        output[0] = ((Char*)&json_len_be)[0];
                        output[1] = ((Char*)&json_len_be)[1];
                        output[2] = ((Char*)&json_len_be)[2];
                        output[3] = ((Char*)&json_len_be)[3];
#endif

                        Grunt_Survey *gs =(Grunt_Survey *) htLookup(&grunt_survey_map,
                                                                    hashInteger((U64)c.id));
                        if (gs == NULL) {
                            gs = calloc(1, sizeof(*gs));
                            htInsert(&grunt_survey_map, hashInteger((U64)c.id), (U64)gs);
                        }
                        gs->milli_last = timeMilli();
                        gs->id = c.id;
                        gs->dispatcher_socket = c.reqin.dispatcher_socket;

                        Sint fd = (Sint)socket_map.keys[i];
                        ssize_t nsent = write(fd, output, output_len);

                        if (nsent == -1) {
                            if (errno == EAGAIN || errno == EWOULDBLOCK) {
                                Output_Resume or = {0};
                                or.fd = fd;
                                or.buffer = output;
                                or.buffer_pos = 0;
                                or.buffer_len = output_len;

                                Input_Output *io = (Input_Output *)htLookup(&io_map, (U64)fd);
                                sbufAdd(io->ors, or);

                                struct epoll_event event = {.data.fd = fd,
                                                            .events = EPOLLIN | EPOLLOUT};
                                if (epoll_ctl(epoll_fd, EPOLL_CTL_MOD, fd, &event) < 0) {
                                    perror("epoll_ctl EPOLL_CTL_MOD");
                                    exit(-1);
                                }
                            } else {
                                perror("write() failed");
                                exit(-1);
                            }
                        } else if ((Size)nsent < output_len) {
                            Output_Resume or = {0};
                            or.fd = fd;
                            or.buffer = output;
                            or.buffer_pos = (Size)nsent;
                            or.buffer_len = output_len;

                            Input_Output *io = (Input_Output *)htLookup(&io_map, (U64)fd);
                            sbufAdd(io->ors, or);

                            struct epoll_event event = {.data.fd = fd,
                                                        .events = EPOLLIN | EPOLLOUT};
                            if (epoll_ctl(epoll_fd, EPOLL_CTL_MOD, fd, &event) < 0) {
                                perror("epoll_ctl EPOLL_CTL_MOD");
                                exit(-1);
                            }
                        } else {
                            struct epoll_event event = {.data.fd = fd,
                                                        .events = EPOLLIN};
                            if (epoll_ctl(epoll_fd, EPOLL_CTL_MOD,
                                          fd, &event) < 0) {
                                perror("epoll_ctl EPOLL_CTL_MOD");
                                exit(-1);
                            }

                            sbufDelete(output);
                        }
                    }
                }
            } else if (c.kind == Command_RESPONSE_EXTERNAL) {
                if (htLookup(&commands_pending, hashInteger((U64)c.id))) {
                    htRemove(&commands_pending, hashInteger((U64)c.id));

                    Sint fd = c.resex.dispatcher_socket;

                    Char *output = NULL;

                    sbufPrint(output, "    ");
                    sbufPrint(output, "{\n\"id\": %d", c.id);
                    sbufPrint(output, ",\n\"ip\": [");
                    for (Size k = 0; k < sbufElemin(c.resex.ips); k++) {
                        sbufPrint(output, "\"%s\"", c.resex.ips[k]);
                        if (k < sbufElemin(c.resex.ips) - 1) {
                            sbufPrint(output, ",");
                        }
                    }
                    sbufPrint(output, "]");
                    sbufPrint(output, "\n}");
                    Size output_len = strlen(output);

                    printf("Sending: RESPONSE EXTERNAL:\n%s\n",
                           cJSON_Print(cJSON_Parse(output + 4)));

#if defined(ENDIAN_LITTLE)
                    U32 json_len = (U32)output_len - 4;
                    U32 json_len_be = swap_endian(json_len);
                    output[0] = ((Char*)&json_len_be)[0];
                    output[1] = ((Char*)&json_len_be)[1];
                    output[2] = ((Char*)&json_len_be)[2];
                    output[3] = ((Char*)&json_len_be)[3];
#endif

                    ssize_t nsent = write(fd, output, output_len);

                    if (nsent == -1) {
                        if (errno == EAGAIN || errno == EWOULDBLOCK) {
                            Output_Resume or = {0};
                            or.fd = fd;
                            or.buffer = output;
                            or.buffer_pos = 0;
                            or.buffer_len = output_len;

                            Input_Output *io = (Input_Output *)htLookup(&io_map, (U64)fd);
                            sbufAdd(io->ors, or);

                            struct epoll_event event = {.data.fd = fd,
                                                        .events = EPOLLIN | EPOLLOUT};
                            if (epoll_ctl(epoll_fd, EPOLL_CTL_MOD, fd, &event) < 0) {
                                perror("epoll_ctl EPOLL_CTL_MOD");
                                exit(-1);
                            }
                        } else {
                            perror("write() failed");
                            exit(-1);
                        }
                    } else if ((Size)nsent < output_len) {
                        Output_Resume or = {0};
                        or.fd = fd;
                        or.buffer = output;
                        or.buffer_pos = (Size)nsent;
                        or.buffer_len = output_len;

                        Input_Output *io = (Input_Output *)htLookup(&io_map, (U64)fd);
                        sbufAdd(io->ors, or);

                        struct epoll_event event = {.data.fd = fd,
                                                    .events = EPOLLIN | EPOLLOUT};
                        if (epoll_ctl(epoll_fd, EPOLL_CTL_MOD, fd, &event) < 0) {
                            perror("epoll_ctl EPOLL_CTL_MOD");
                            exit(-1);
                        }
                    } else {
                        struct epoll_event event = {.data.fd = fd,
                                                    .events = EPOLLIN};
                        if (epoll_ctl(epoll_fd, EPOLL_CTL_MOD,
                                      fd, &event) < 0) {
                            perror("epoll_ctl EPOLL_CTL_MOD");
                            exit(-1);
                        }

                        sbufDelete(output);
                    }
                }
            }
        }

        sbufDelete(commands);
    }

}
